/* Sentinel - IRC Statistical and Operator Services
** log.c - Misc. utility functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: log.c,v 1.7 2003/08/08 13:57:40 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"
#include "token.h"

#ifdef SPLITSERV
FILE *split_log = NULL;
#endif
FILE *main_log = NULL;

#ifdef IO_LOG
FILE *io_log = NULL;
#endif

#if defined(MONSERV) && defined(MON_DRONES)
FILE *drone_log = NULL;
#endif

FILE *spy_log = NULL;

struct Token log_table[] =
  {
    { "NONE", L_NONE },
    { "DEBUG", L_DEBUG },
    { "INFO", L_INFO },
    { "NOTICE", L_NOTICE },
    { "WARNING", L_WARNING },
    { "ERR", L_ERR },
    { "CRIT", L_CRIT },
    { "ALERT", L_ALERT },
    { "EMERG", L_EMERG },
    { NULL, TOKEN_ERROR }
  };

/* Generic logging routine */
void slog(int level, char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE], fmtime[LINE];
  time_t tmp = time(NULL);

  /* Logging is done if the level of this log is greater than or
  ** equal to the level specified in the conf. (like syslog)
  */
  if (level < settings.log_level)
    return ;

  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);

  strftime(fmtime, LINE, TIMEFMT, localtime(&tmp));

  if (main_log == NULL)
  {
    main_log = fopen(LOGFILE, "a");
    if (main_log == NULL)
      return ;
  }

  fprintf(main_log, "(%s) %s\n", fmtime, buf);

  if (IsDebug())
  {
    fprintf(stderr, "(%s) %s\n", fmtime, buf);
  }

  va_end(ap);
}

#if defined(MONSERV) && defined(MON_DRONES)
void dronelog(char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE];
  time_t tmp = time(NULL);

  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);

  if (drone_log == NULL)
  {
    drone_log = fopen(DRONELOG, "a");
    if (drone_log == NULL)
      return ;
  }

  fprintf(drone_log, "(%ld) %s\n", (long)tmp, buf);

  if (IsDebug())
  {
    fprintf(stderr, "(%ld) %s\n", (long)tmp, buf);
  }

  va_end(ap);
}
#endif

void spylog(char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE], fmtime[LINE];
  time_t tmp = time(NULL);

  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);

  strftime(fmtime, LINE, TIMEFMT, localtime(&tmp));

  if (spy_log == NULL)
  {
    spy_log = fopen(SPYLOG, "a");
    if (spy_log == NULL)
      return ;
  }

  fprintf(spy_log, "(%s) %s\n", fmtime, buf);

  if (IsSpyWallops())
  {
    wallops(settings.name, "%s", buf);
  }

  if (IsDebug())
  {
    fprintf(stderr, "(%s) %s\n", fmtime, buf);
  }

  va_end(ap);
}

#ifdef IO_LOG
void iolog(char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE];

  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);

  if (io_log == NULL)
  {
    io_log = fopen(IOLOGFILE, "a");
    if (io_log == NULL)
      return ;
  }

  fprintf(io_log, "%s\n", buf);

  va_end(ap);
}
#endif

#ifdef SPLITSERV 
/* SplitServ logging routine */
void splog(char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE], fmtime[LINE];
  time_t tmp = time(NULL);

  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);

  strftime(fmtime, LINE, TIMEFMT, localtime(&tmp));

  if (split_log == NULL)
  {
    split_log = fopen(SPLITLOG, "a");
    if (split_log == NULL)
      return ;
  }

  fprintf(split_log, "(%s) %s\n", fmtime, buf);

  if (IsDebug())
  {
    fprintf(stderr, "(%s) %s\n", fmtime, buf);
  }

  va_end(ap);
}
#endif

/* Rotate the logfile, called from chk() */
void ResetLogs()
{
  char tmp[BUFSIZE];
  time_t t = time(NULL);
  struct tm *ltm;

  /* Fix the date stamp so it's the logs of THAT day, not the previous
  ** day with today's date stamp on it
  */
  t = t - 86400;
  /* It's MUCH easier to manipulate this than all the vars inside the
  ** struct
  */
  ltm = localtime(&t);

  strftime(tmp, BUFSIZE - 1, BACKUPLOGFMT, ltm);
  tmp[BUFSIZE - 1] = '\0';
  if (main_log != NULL)
  {
    fclose(main_log);
    main_log = NULL;
  }
  rename(LOGFILE, tmp);
  /* log will reopen the log file if it's closed */
  slog(L_CRIT, "Started fresh logfile.");

#ifdef SPLITSERV

  strftime(tmp, BUFSIZE, BACKUPSPLOGFMT, ltm);
  tmp[BUFSIZE - 1] = '\0';
  if (split_log != NULL)
  {
    fclose(split_log);
    split_log = NULL;
  }
  rename(SPLITLOG, tmp);
  splog("Started fresh logfile.");
#endif
#if defined(MONSERV) && defined(MON_DRONES)

  strftime(tmp, BUFSIZE, BACKUPDRONEFMT, ltm);
  tmp[BUFSIZE - 1] = '\0';
  if (drone_log != NULL)
  {
    fclose(drone_log);
    drone_log = NULL;
  }
  rename(DRONELOG, tmp);
  dronelog("Started fresh logfile.");
#endif

  strftime(tmp, BUFSIZE, BACKUPSPYFMT, ltm);
  tmp[BUFSIZE - 1] = '\0';
  if (spy_log != NULL)
  {
    fclose(spy_log);
    spy_log = NULL;
  }
  rename(SPYLOG, tmp);
  spylog("Started fresh logfile.");
}

void close_logfiles(void)
{
  if (main_log != NULL)
  {
    fclose(main_log);
    main_log = NULL;
  }
#ifdef SPLITSERV
  if (split_log != NULL)
  {
    fclose(split_log);
    split_log = NULL;
  }
#endif
#ifdef IO_LOG
  if (io_log != NULL)
  {
    fclose(io_log);
    io_log = NULL;
  }
#endif
#if defined(MONSERV) && defined(MON_DRONES)
  if (drone_log != NULL)
  {
    fclose(drone_log);
    drone_log = NULL;
  }
#endif
  if (spy_log != NULL)
  {
    fclose(spy_log);
    spy_log = NULL;
  }
}

