/* Sentinel - IRC Statistical and Operator Services
** users.c - Functions to track users state
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: users.c,v 1.153 2003/08/07 01:42:15 wcampbel Exp $
*/

#include <assert.h>
#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"
#include "sql.h"

static struct User *new_user(char *);
static void add_user_to_hash_table(char *, struct User *);
static void del_user_from_hash_table(struct User *);

struct User *userlist[U_TABLE_SIZE];

static void add_user_to_hash_table(char *nick, struct User *u)
{
  u->hash = HASH((unsigned char *) nick, U_TABLE_SIZE);
  u->next = userlist[u->hash];
  userlist[u->hash] = (void *)u;
}

static void del_user_from_hash_table(struct User *u)
{
  struct User *tmp, *prev = NULL;

  for (tmp = userlist[u->hash]; tmp; tmp = tmp->next)
  {
    if (tmp == u)
    {
      if (prev)
        prev->next = tmp->next;
      else
      {
        userlist[u->hash] = tmp->next;
      }
      tmp->next = NULL;
      return ;
    }
    prev = tmp;
  }
}

static struct User *new_user(char *nick)
{
  struct User *u;

  u = smalloc(sizeof(struct User));
  if (!nick)
    nick = "";
  strlcpy(u->nick, nick, MAXNICK + 1);
  add_user_to_hash_table(nick, u);

  return u;
}

struct User *AddUser(char *nick, char *user, char *host, struct Server *server,
                           char *gecos)
{
#ifdef HAVE_MYSQL
#ifdef TRACK_GECOS
  char *egecos = NULL;
#endif
  char *eusername = NULL, *enick = NULL;
#endif
  struct User *u;

  debug("AddUser(): %s (%s@%s) -> %s", nick, user, host,
        server->name);

  u = finduser(nick);
  if (u)
  {
    slog(L_WARNING, "trying to add a user that already exists? (%s)",
         nick);
    return u;
  }

  u = new_user(nick);
  strlcpy(u->hostname, host, MAXHOST);
  strlcpy(u->username, user, MAXUSER + 1);
  u->server = server;
#ifdef HAVE_MYSQL
#ifdef TRACK_GECOS
  egecos = db_escape(gecos);
  eusername = db_escape(u->username);
  enick = db_escape(nick);
  db_query("INSERT INTO user (nick, realname, username, hostname, server, connecttime) VALUES('%s', '%s', '%s', '%s', '%s', NOW())", enick, egecos, eusername, u->hostname, server->name);
  free(egecos);
  free(eusername);
  free(enick);
#else
  eusername = db_escape(u->username);
  enick = db_escape(nick);
  db_query("INSERT INTO user (nick, username, hostname, server, connecttime) VALUES('%s', '%s', '%s', '%s', NOW())", enick, eusername, u->hostname, server->name);
  free(eusername);
  free(enick);
#endif
#endif

  u->t_flood = time(NULL);
  u->tld = AddTLD(u);
#ifdef TRACK_GECOS

  if (gecos != NULL)
  {
    u->gecos = strdup(gecos);
    memcounts.gecos++;
    memcounts.gecosmem += strlen(gecos) + 1;
  }
#endif

  if (!u->server)
    slog(L_WARNING, "Unable to find server for user %s (%s?)", nick,
         server);
  else
    IncreaseUsers(u->server);

  netinfo.connections++;
  daily.connections++;

#ifdef MONSERV

  AddUserToHashes(u);
#ifdef MON_DRONES

  CheckDrones(u);
  if (gecos != NULL)
    CheckGecosDrones(u, gecos);
#endif
#ifdef TRACK_CTCPVERSION

  if (!mn_delayed)
    sts(":%s PRIVMSG %s :\001VERSION\001", MonServ.nick, u->nick);
#endif
#endif

  return u;
}

void DelUser(char *nick)
{
  struct User *u = finduser(nick);
  dlink_node *dl;
  dlink_node *tdl;
  struct ChanUser *cu;
#ifdef HAVE_MYSQL
  char *enick;
#endif

  debug("DelUser(%s)", nick);

  if (!u)
  {
    slog(L_WARNING, "DelUser(%s) failed!", nick);
    return ;
  }

  del_user_from_hash_table(u);

  DecreaseUsers(u->server);

  if (u->tld)
    u->tld->users--;
  if (IsAway(u))
  {
    netinfo.away--;
    u->server->away--;
  }
  if (IsOper(u))
  {
    netinfo.opers--;
    u->server->opers--;
    DelOperFromList(&global_oper_list, u);
  }
  if (IsInvis(u))
  {
    if (netinfo.invis)
      netinfo.invis--;
    if (u->server->invis)
      u->server->invis--;
  }
  if (u->myuser)
  {
    u->myuser->flags &= ~MFLAG_ISON;
    /* If the myuser struct is marked as a dying
    ** one, and they aren't logged in via DCC,
    ** remove it.
    */
    if (IsOperDying(u->myuser) && !IsOperDCC(u->myuser))
    {
      tdl = dlink_find_delete(u->myuser, &myudying);
      dlink_free(tdl);
#ifdef HAVE_LIBCRYPTO

      if (u->myuser->response != NULL)
        free(u->myuser->response);
      if (u->myuser->keyfile != NULL)
        RSA_free(u->myuser->keyfile);
#endif

      free(u->myuser);
      memcounts.myuser--;
    }
  }

  /* Remove the user from each channel */
  DLINK_FOREACH_SAFE(dl, tdl, u->channels.head)
  {
    cu = (struct ChanUser *) dl->data;
    DelUserFromChannel(cu->chan, u);
    /* The ChanUser is free'd as a side effect of
    ** DelUserFromChannel()
    */
    dlink_delete(dl, &u->channels);
    dlink_free(dl);
  }

  /* Debug asserts */
  assert(u->tld->users >= 0);
  assert(u->server->opers >= 0);
  assert(netinfo.opers >= 0);
  assert(netinfo.users >= 0);

#ifdef TRACK_GECOS

  if (u->gecos != NULL)
  {
    memcounts.gecos--;
    memcounts.gecosmem -= strlen(u->gecos) + 1;
    free(u->gecos);
  }
#endif

#ifdef MONSERV
  DelUserFromHashes(u);
#ifdef MON_DRONES

  RemoveDrone(u);
#endif
#endif

#ifdef TRACK_CTCPVERSION

  if (u->version != NULL)
  {
    memcounts.cversion--;
    memcounts.cversionmem -= strlen(u->version) + 1;
    free(u->version);
  }
#endif

#ifdef HAVE_MYSQL
  enick = db_escape(nick);
  db_query("DELETE FROM user WHERE nick='%s'", enick);
  free(enick);
#endif

  free(u);
}

#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
void Change_User(struct User *u, char *newnick, int issvs)
#else
void Change_User(struct User *u, char *newnick)
#endif
{
#ifdef HAVE_MYSQL
  char *enick, *ennick;
#endif

  debug("Change_User(%s, %s)", u->nick, newnick);

#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)

  spy("%s Change by %s to %s", issvs ? "SVSNICK" : "NICK", u->nick,
      newnick);
  if (issvs)
    sendto_dcc_flag(CF_SERVICES, "SVSNICK Change by %s to %s",
                    u->nick, newnick);
  else
    sendto_dcc_flag(CF_USER, "NICK Change by %s to %s",
                    u->nick, newnick);
#else

  spy("NICK Change by %s to %s", u->nick, newnick);
  sendto_dcc_flag(CF_USER, "NICK Change by %s to %s", u->nick, newnick);
#endif

#ifdef HAVE_MYSQL
  enick = db_escape(u->nick);
  ennick = db_escape(newnick);
  db_query("UPDATE user SET nick=\"%s\" WHERE nick='%s'",
           ennick, enick);
  db_query("UPDATE ison SET nick=\"%s\" WHERE nick='%s'",
           ennick, enick);
  free(enick);
  free(ennick);
#endif

  del_user_from_hash_table(u);

  strlcpy(u->nick, newnick, MAXNICK + 1);

  add_user_to_hash_table(u->nick, u);
}

#ifdef SUPPORT_HOSTSERV
void Change_Host(struct User *u, char *newhost)
{
#ifdef HAVE_MYSQL
  char *enick;
#endif

  debug("Change_Host(%s, %s)", u->hostname, newhost);

  spy("SVSHOST Change by %s to %s", u->hostname, newhost);
  sendto_dcc_flag(CF_SERVICES, "SVSHOST Change By %s to %s",
                  u->hostname, newhost);
#ifdef HAVE_MYSQL
  enick = db_escape(u->nick);
  db_query("UPDATE user SET hostname=\"%s\" WHERE nick='%s'",
           newhost, enick);
  free(enick);
#endif

  strlcpy(u->hostname, newhost, MAXHOST);
  u->tld = AddTLD(u);
}

void Change_Ident(struct User *u, char *newident)
{
#ifdef HAVE_MYSQL
  char *enick;
#endif

  debug("Change_Host(%s, %s)", u->hostname, newident);

  spy("SVSIDENt Change by %s to %s", u->username, newident);
  sendto_dcc_flag(CF_SERVICES, "SVSIDENT Change By %s to %s",
                  u->username, newident);
#ifdef HAVE_MYSQL
  enick = db_escape(u->nick);
  db_query("UPDATE user SET username=\"%s\" WHERE nick='%s'",
           newident, u->nick);
  free(enick);
#endif

  strlcpy(u->username, newident, MAXUSER + 1);
}
#endif

struct User *finduser(char *nick)
{
  struct User *u;

  u = userlist[HASH((unsigned char *) nick, U_TABLE_SIZE)];
  while (u && (irccasecmp(u->nick, nick) != 0))
    u = u->next;

  debug("finduser(%s) -> %s", nick, (u) ? u->nick : "NOTFOUND");

  return u;
}

void init_user_hash()
{
  int i;
  struct User *u, *prev;

  for (i = 0; i < U_TABLE_SIZE; i++)
  {
    u = userlist[i];
    while (u != NULL)
    {
      prev = u->next;
      debug("init_user_hash() deleting %s", u->nick);
      free(u);

      u = prev;
    }
    userlist[i] = NULL;
  }
#ifdef HAVE_MYSQL
  db_query("DELETE FROM user");
#endif
}

void UserMode(struct User *u, char *modes)
{
#ifdef HAVE_MYSQL
  char *enick;
#endif
  int add = 0;

  if (!u)
  {
    slog(L_WARNING, "Changing modes for unknown user");
    return ;
  }

  while (*modes != '\0')
  {
    if (*modes == '+')
      add = 1;
    if (*modes == '-')
      add = 0;
#ifdef HAVE_MYSQL
    enick = db_escape(u->nick);
    if (!(*modes == '+' || *modes == '-'))
    {
    if (IsLower(*modes))
      if (add)
        db_query("UPDATE user SET m_l%c=\"Y\" WHERE nick='%s'", (char) *modes, enick);
      else
        db_query("UPDATE user SET m_l%c=\"N\" WHERE nick='%s'", (char) *modes, enick);
    else
      if (add)
        db_query("UPDATE user SET m_u%c=\"Y\" WHERE nick='%s'", (char) *modes, enick);
      else
        db_query("UPDATE user SET m_u%c=\"N\" WHERE nick='%s'", (char) *modes, enick);
    }
    free(enick);
#endif
    switch (*modes)
    {
      case '+':
        add
          = 1;
        break;
      case '-':
        add
          = 0;
        break;
      case 'o':
        if (add
           )
        {
          if (!IsOper(u))
          {
            set_oper(u, UFLAG_ISOPER, OPER_GLOBAL);

            alert("New Oper %s on %s", u->nick,
                  u->server->name);
            sendto_dcc_flag(CF_OPER,
                            "New Oper %s on %s", u->nick,
                            u->server->name);
            debug("%s is now an IRCop.", u->nick);
          }
        }
        else
        {
          if (IsOper(u))
          {
            u->flags &= ~UFLAG_ISOPER;
            DelOperFromList(&global_oper_list, u);
            netinfo.opers--;
            u->server->opers--;
#ifdef OPER_MODES

            u->operflags = 0;
#endif

            alert("Deoper by %s on %s", u->nick,
                  u->server->name);
            sendto_dcc_flag(CF_OPER,
                            "Deoper by %s on %s", u->nick,
                            u->server->name);
            debug("%s is no longer an IRCop.",
                  u->nick);
            assert(netinfo.opers >= 0);
            assert(u->server->opers >= 0);
          }
        }
        break;
#ifdef IRCD_2_10
        /* IRCnet propegates AWAY by setting usermode +a */
      case 'a':
        if (add
           )
        {
          if (!IsAway(u))
          {
            u->flags |= UFLAG_ISAWAY;
            IncreaseAway(u->server);
          }
        }
        else
        {
          if (IsAway(u))
          {
            u->flags &= ~UFLAG_ISAWAY;
            netinfo.away--;
            u->server->away--;
          }
          assert(netinfo.away >= 0);
        }
        break;
#endif
#ifdef OPER_MODES
#ifdef IRCD_BAHAMUT

      case 'A':
#else

      case 'a':
#endif

        if (add
           )
          u->operflags |= OPER_ADMIN;
        else
          u->operflags &= ~OPER_ADMIN;
        if (add
           )
        {
          alert("New Server Admin %s on %s",
                u->nick, u->server->name);
          sendto_dcc_flag(CF_OPER,
                          "New Server Admin %s on %s",
                          u->nick, u->server->name);
        }
        else
        {
          alert("Server Admin Status Removal %s "
                "on %s", u->nick,
                u->server->name);
          sendto_dcc_flag(CF_OPER,
                          "New Server Admin %s on %s",
                          u->nick, u->server->name);
        }
        break;
#ifdef IRCD_BAHAMUT

      case 'a':
        if (add
           )
          u->operflags |= OPER_SVSADMIN;
        else
          u->operflags &= ~OPER_SVSADMIN;
        if (add
           )
        {
          alert("New Services Admin %s on %s",
                u->nick, u->server->name);
          sendto_dcc_flag(CF_OPER,
                          "New Services Admin %s on %s",
                          u->nick, u->server->name);
        }
        else
        {
          alert("Services Admin Status Removal "
                "%s on %s", u->nick,
                u->server->name);
          sendto_dcc_flag(CF_OPER,
                          "Services Admin Status Removal "
                          "%s on %s", u->nick,
                          u->server->name);
        }
        break;
#endif
#endif

      case 'O':
        /* Ignore locals */
        break;

      case 'i':
        if (add
           )
        {
          u->flags |= UFLAG_INVIS;
          netinfo.invis++;
          u->server->invis++;
        }
        else
        {
          u->flags &= ~UFLAG_INVIS;
          if (netinfo.invis)
            netinfo.invis--;
          if (u->server->invis)
            u->server->invis--;
        }
        break;

      default:
        break;
    }
    modes++;
  }
}

void IncreaseUsers(struct Server *s)
{
  time_t current = time(NULL);

  debug("Increasing UserCount for %s", s->name);
  s->users++;
  netinfo.users++;
  if (s->users > s->stats->maxusers)
  {
    s->stats->maxusers = s->users;
    s->stats->t_maxusers = current;
  }
  if (netinfo.users > netinfo.maxusers)
  {
    netinfo.maxusers = netinfo.users;
    netinfo.t_maxusers = current;
  }
  if (netinfo.users > daily.users)
  {
    daily.users = netinfo.users;
    daily.t_users = current;
  }
}

